#include "CLogging.h"
#include "common.h"
#include <time.h>

bool CLogging::m_globalEnableLogging = true;	
bool CLogging::m_globalEnableLogToFile = true;
unsigned int CLogging::m_globalDebugLevel = CLogging::DEBUG_LEVEL_ABOVE_AVERAGE;
std::string CLogging::m_logFile = "";
struct tm CLogging::m_currentLogDate;

void CLogging::LogToStdOut(const char * buffer, const unsigned short length) {
	#if defined (_MSC_VER) || defined (__GNUC__) 
	printf("%s", buffer);
	printf("\n");
	fflush(stdout);
	#endif
}

void CLogging::LogToFile(const char * buffer, const unsigned short length) {
	// Check if logging to file is enabled
	if (!CLogging::m_globalEnableLogToFile) {
		return;
	}

	if (CLogging::m_logFile.empty()) {
		return;
	}

	CFileIo::AppendFile(CLogging::m_logFile, buffer, length);
	std::string endofline = "\n";
	CFileIo::AppendFile(CLogging::m_logFile, endofline.c_str(), endofline.size());
}

void CLogging::LogMessage(const unsigned int logLevel, std::string logType, const char * lpszText, va_list argList) {
	// Has the logging system been enabled
	if (!CLogging::m_globalEnableLogging) {
		return;
	}

	// Check logging Level
	if (logLevel > CLogging::m_globalDebugLevel) {
		return; // Too low to report.
	}

	// Check fi there is anything to print
	if (lpszText == NULL) {
		return;
	}

	char * buffer = (char *)malloc(sizeof(char) * CLOGGING_MAX_LINE_LENGTH);
	if (buffer == NULL) {
		return; // Out of Memory
	}

	unsigned short length = 0;
	if (argList == NULL) {
		length += snprintf(buffer + length, CLOGGING_MAX_LINE_LENGTH - length, "argList is NULL");
	}
	else {
		// Output the format
		length += snprintf(buffer, CLOGGING_MAX_LINE_LENGTH, "%s: ", logType.c_str());
		length += vsnprintf(buffer + length, CLOGGING_MAX_LINE_LENGTH - length, lpszText, argList);
	}

	// Send the message
	CLogging::LogSend(logLevel, buffer, NULL);

	SAFE_FREE(buffer);
}

void CLogging::LogSend(const unsigned int logLevel, const char * lpszText, va_list argList) {
	// Has the logging system been enabled
	if (!CLogging::m_globalEnableLogging) {
		return;
	}

	// Check logging Level
	if (logLevel > CLogging::m_globalDebugLevel) {
		return; // Too low to report.
	}

	// Check fi there is anything to print
	if (lpszText == NULL) {
		return;
	}

	// Add the time
	char  * buffer = (char *)malloc(sizeof(char) * CLOGGING_MAX_LINE_LENGTH);
	if (buffer == NULL) {
		return;	// Out of memory
	}

	unsigned short length = snprintf(buffer, CLOGGING_MAX_LINE_LENGTH, "[%ld]", time(0));

	if (argList == NULL) {
		length += snprintf(buffer + length, CLOGGING_MAX_LINE_LENGTH - length, "%s", lpszText);
	}
	else {
		length += vsnprintf(buffer + length, CLOGGING_MAX_LINE_LENGTH - length, lpszText, argList);
	}

	// Send the message
	#if defined(_MSC_VER) || defined(__GNUC__)
	CLogging::LogToStdOut(buffer, length);
	#endif // WIN32

	CLogging::LogToFile(buffer, length);


	// Free memory
	SAFE_FREE(buffer);
}


void CLogging::Error(const char * lpszText, ...) {
	// Has the logging system been enabled 
	if (!CLogging::m_globalEnableLogging) {
		return;
	}

	// Check to see if any message was sent. 
	if (lpszText == NULL) {
		return;
	}

	// Initialize variable argument list
	va_list argList;
	va_start(argList, lpszText);
	CLogging::LogMessage(CLogging::DEBUG_LEVEL_CRITICAL, "Error", lpszText, argList);
	va_end(argList);
}

void CLogging::Fyi(const char * lpszText, ...) {
	// Has the logging system been enabled 
	if (!CLogging::m_globalEnableLogging) {
		return;
	}

	// Check to see if any message was sent. 
	if (lpszText == NULL) {
		return;
	}

	// Initialize variable argument list
	va_list argList;
	va_start(argList, lpszText);
	CLogging::LogMessage(CLogging::DEBUG_LEVEL_CRITICAL, "FYI", lpszText, argList);
	va_end(argList);
}

void CLogging::Debug(const unsigned int logLevel, const char * lpszText, ...) {
	// Has the logging system been enabled 
	if (!CLogging::m_globalEnableLogging) {
		return;
	}

	// Check to see if any message was sent. 
	if (lpszText == NULL) {
		return;
	}
	
	// Generate type
	char type[100];
	snprintf(type, 100, "Debug(%u)", logLevel);

	// Initialize variable argument list
	va_list argList;
	va_start(argList, lpszText);
	CLogging::LogMessage(logLevel, type, lpszText, argList);
	va_end(argList);
}

void CLogging::LogBuffer(const unsigned int logLevel, std::string buffer) {
	// Has the logging system been enabled 
	if (!CLogging::m_globalEnableLogging) {
		return;
	}

	// Check logging Level
	if (logLevel > CLogging::m_globalDebugLevel) {
		return; // Too low to report.
	}

	char type[32];
	snprintf(type, 32, "Debug(%u): ", logLevel);

	char timestamp[32];
	snprintf(timestamp, 32, "[%ld]", time(0));

	std::string toLog = "";
	toLog.append(timestamp);
	toLog.append(type);
	toLog.append(buffer);

	CLogging::LogToStdOut(toLog.c_str(), toLog.size());
	CLogging::LogToFile(toLog.c_str(), toLog.size());
}

void CLogging::SetCurrentLogDate() {
	time_t currentTime = time(0);
	struct tm * temp = localtime(&currentTime);
	CLogging::m_currentLogDate = *temp;
}

void CLogging::LogFileManagement() {
	// First check the time
	time_t currentTime = time(0);
	struct tm * temp = localtime(&currentTime);

	bool changeLogFile = false;

	if (temp->tm_year != CLogging::m_currentLogDate.tm_year ||
		temp->tm_mon != CLogging::m_currentLogDate.tm_mon ||
		temp->tm_mday != CLogging::m_currentLogDate.tm_mday)
	{
		// New logging date
		changeLogFile = true;
	}


	if (changeLogFile) {
		// New date
		CLogging::SetCurrentLogDate();
		// Get the newest file name and store
		char logFilename[1024];
		snprintf(logFilename, 1024, "%s/logs/%d%02d%d_log.txt", CFileIo::GetWorkingDirectory().c_str(), CLogging::m_currentLogDate.tm_year + 1900, CLogging::m_currentLogDate.tm_mon + 1, CLogging::m_currentLogDate.tm_mday);
		CLogging::SetLoggingFile(logFilename);

		// Check for clean up
		if (CFileIo::GetNumberOfFiles("/logs") > 4) {
			// Get the oldest file and delete
			std::string oldestFileName = CFileIo::GetOldestFileInDirectory(CFileIo::GetWorkingDirectory(), "logs", "*.txt");
			if (!oldestFileName.empty()) {
				CFileIo::RemoveFile(oldestFileName);
			}
		}
	}
}
