#include "common.h"

#include <cctype>
#include <vector>

#ifdef _MSC_VER
#include <direct.h>
#include <windows.h>
#elif defined __GNUC__
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <limits.h>
#include <unistd.h>
#include <dlfcn.h>
#endif 


unsigned char CUtilities::SetBit(unsigned char data, unsigned char bit, bool value) {
	data ^= (-(value ? 1 : 0) ^ data) & (1 << bit);
	return data;
}

bool CUtilities::GetBit(unsigned char data, unsigned char bit) {
	return ((data >> bit) & 1);
}

bool CUtilities::IsNumber(std::string s) {
	for (size_t i = 0; i < s.size(); i++) {
		// Account for negative numbers
		if (i == 0 && s[i] == '-') {
			continue;
		}

		if (!std::isdigit(s[i])) {
			return false;
		}
	}

	return true;
}

bool CUtilities::ValidateIPAddress(std::string ipAddress) {
	if (ipAddress.size() < 7 || ipAddress.size() > 15) {
		return false; // Invalid length
	}

	char tokens[] = ".";
	int quadCount = 0;

	char buffer[100];
	memcpy(buffer, ipAddress.c_str(), ipAddress.size());
	buffer[ipAddress.size()] = 0;


	char * quad = strtok(buffer, tokens);
	while (quad != NULL) {
		quadCount++;

		// Check that the quad is between 0 and 255
		if (!CUtilities::IsNumber(quad)) {
			return false;
		}

		int num = atoi(quad);
		if (num < 0 || num > 255) {
			return false; // Invalid quad value
		}

		if ((quadCount == 1 && num == 0) ||
			(quadCount == 4 && num == 0)) {
			return false; // IP address cannot be of the form: 0.xxx.xxx.xxx or xxx.xxx.xxx.0
		}

		quad = strtok(NULL, tokens);
	}

	if (quadCount != 4) {
		return false;
	}

	return true;
}


std::string CConvert::ToCamelCase(std::string value)
{
	// Remove all spaces or _ 
	// Set first letter after space or _ to uppercase
	std::string result = "";
	bool firstToken = true;
	char * token = strtok((char*)value.c_str(), " _");
	while (token) {
		if (firstToken) {
			firstToken = false;
		}
		else {
			token[0] = toupper(token[0]);
		}

		result += std::string(token);

		token = strtok(NULL, " _");
	}

	return result;
}

std::string CConvert::UndoCamelCase(std::string value, char spaceOrSymbol, bool allLowerCase)
{
	// Flag for number handling
	// Example: proprietaryObjectType500 = proprietary_object_type_500 if using _ as a symbol
	bool firstNumber = true;

	// Cycle through each char
	std::string result = "";

	for (size_t i = 0; i < value.size(); i++) {
		if (isupper(value[i])) {
			result.push_back(spaceOrSymbol);
			if (allLowerCase) {
				result.push_back(tolower(value[i]));
			}
			else {
				result.push_back(value[i]);
			}

			// Reset flag if followed by a capital
			if (!firstNumber) {
				firstNumber = true;
			}
		}
		else if (isdigit(value[i])) {
			if (firstNumber) {
				result.push_back(spaceOrSymbol);
				firstNumber = false;
			}
			result.push_back(value[i]);
		}
		else {
			result.push_back(value[i]);
		}
	}

	return result;
}

bool CConvert::IPAddressToByteString(std::string ipAddress, uint8_t * byteString, const uint8_t maxByteStringLength) {
	// Check if the ip address is valid
	if (!CUtilities::ValidateIPAddress(ipAddress)) {
		return false;
	}

	// Verify that there is enough bytes to write (need at least 4)
	if (maxByteStringLength < 4) {
		return false;
	}

	// Setup the parameters needed to extract the byte values from the ip address
	int byte0, byte1, byte2, byte3;
	char dot;

	// Extract the values using istringstream
	std::istringstream s(ipAddress);
	s >> byte0 >> dot >> byte1 >> dot >> byte2 >> dot >> byte3;

	// Store values in the byteString parameter
	byteString[0] = byte0;
	byteString[1] = byte1;
	byteString[2] = byte2;
	byteString[3] = byte3;

	return true;
}

bool CFileIo::MkDir(std::string path)
{
	int status;

	#ifdef _MSC_VER
	status = _mkdir(path.c_str());
	#elif defined __GNUC__
	status = mkdir(path.c_str(), S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
	#endif 

	if (status == 0) {
		return true;
	}

	int error = errno;
	if (error == EEXIST) {
		return true;
	}

	return false;
}

bool CFileIo::ExistFile(std::string path)
{
	FILE * file = fopen(path.c_str(), "r");
	if (file == NULL) {
		return false; // Nothing to do.
	}
	fclose(file);

	return true;
}

bool CFileIo::AppendFile(std::string path, const char * buffer, size_t bufferLength)
{
	bool ret = false;
	FILE * stream = fopen(path.c_str(), "a+");
	if (stream != NULL) {
		ret = true;
		if (fwrite(buffer, 1, bufferLength, stream) != bufferLength) {
			ret = false;
		}
		fclose(stream);
	}

	return ret;
}

bool CFileIo::RemoveFile(std::string path)
{
	if (!CFileIo::ExistFile(path)) {
		return true;
	}

	return (remove(path.c_str()));
}

time_t CFileIo::GetFileTime(std::string path)
{
	time_t ret = 0;

	#ifdef _MSC_VER
	struct stat st;
	int ierr = stat(path.c_str(), &st);
	if (ierr != 0) {
		// Error
		return 0;
	}

	ret = st.st_mtime;
	#elif defined __GNUC__
	struct stat st;
	stat(path.c_str(), &st);
	ret = st.st_mtim.tv_sec;
	#endif
	
	return ret;
}

int CFileIo::GetNumberOfFiles(std::string path)
{
	int counter = 0;

	#ifdef _MSC_VER
	WIN32_FIND_DATA ffd;
	HANDLE hFind = INVALID_HANDLE_VALUE;

	// Start iterating over the files in the path directory.
	std::string search = path + std::string("\\*");
	hFind = FindFirstFile(search.c_str(), &ffd);
	if (hFind != INVALID_HANDLE_VALUE)
	{
		do // Managed to locate and create an handle to that folder.
		{
			std::string filename = std::string(ffd.cFileName);
			if (filename.compare(".") != 0 && filename.compare("..") != 0) {
				counter++;
			}
		} while (FindNextFile(hFind, &ffd) == TRUE);
		FindClose(hFind);
	}
	#elif defined __GNUC__
	DIR * dir;
	struct dirent * entry;

	dir = opendir(path.c_str());
	if( dir == NULL ) {
		return 0 ;
	}

	while((entry = readdir(dir)) != NULL ) {
		if( entry->d_type == DT_REG) {
			counter++;
		}
	}
	closedir(dir);
	#endif

	return counter;
}

std::string CFileIo::GetOldestFileInDirectory(std::string path, const char * dir, const char * pattern)
{
	std::string oldestFileName = "";

	#ifdef WIN32
	bool firstFile = true;
	// Search through the files in the given directory and find the older file.

	FILETIME oldestFileTime;
	WIN32_FIND_DATA fileData;
	std::vector<std::string> directoryNames;
	HANDLE h;

	std::string search = path + std::string(dir).append("\\*");

	// First process subdirectories
	h = FindFirstFile(search.c_str(), &fileData);
	if (h != INVALID_HANDLE_VALUE)
	{
		do {
			if (fileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				std::string directory = std::string(fileData.cFileName);
				// Search subdirectory
				if (directory.compare(".") == 0 || directory.compare("..") == 0) {
					continue;
				}
				directoryNames.push_back(directory);
			}
		} while (FindNextFile(h, &fileData));
	}
	directoryNames.push_back(dir);

	// Now search all the subdirectories
	for (size_t i = 0; i < directoryNames.size(); i++) {
		if (directoryNames[i].compare(dir) == 0) {
			search = path + directoryNames[i] + std::string("\\").append(pattern);
		}
		else {
			search = path + std::string(dir).append("\\") + directoryNames[i] + std::string("\\").append(pattern);
		}

		h = FindFirstFile(search.c_str(), &fileData);
		if (h != INVALID_HANDLE_VALUE)
		{
			do {
				if (firstFile) {
					if (directoryNames[i].compare(dir) == 0) {
						oldestFileName = path + directoryNames[i] + std::string("\\").append(fileData.cFileName);
					}
					else {
						oldestFileName = path + std::string(dir).append("\\") + directoryNames[i] + std::string("\\").append(fileData.cFileName);
					}
					oldestFileTime = fileData.ftCreationTime;
					firstFile = false;
				}

				if (CompareFileTime(&oldestFileTime, &fileData.ftCreationTime) > 0) {
					if (directoryNames[i].compare(dir) == 0) {
						oldestFileName = path + directoryNames[i] + std::string("\\").append(fileData.cFileName);
					}
					else {
						oldestFileName = path + std::string(dir).append("\\") + directoryNames[i] + std::string("\\").append(fileData.cFileName);
					}
					oldestFileTime = fileData.ftCreationTime;
				}
			} while (FindNextFile(h, &fileData));
			FindClose(h);
		}	
	}
	#elif defined (__GNUC__)

	std::string search = path + std::string("/") +  std::string(dir);

	DIR * directory;
	struct dirent * entry;

	directory = opendir(search.c_str());
	if( directory == NULL ) {
		return 0 ;
	}

	bool firstFile = true;
	time_t oldestTime = 0;

	while((entry = readdir(directory)) != NULL ) {
		if( entry->d_type == DT_REG) {
			std::string fileName = search + std::string("/") + std::string(entry->d_name);
			time_t fileTime = CFileIo::GetFileTime(fileName);

			if(firstFile) {
				oldestFileName = fileName;
				oldestTime = fileTime;
				firstFile = false;
			} else {
				if(fileTime < oldestTime ) {
					oldestFileName = fileName;
					oldestTime = fileTime;
				}
			}
		}
	}
	closedir(directory);
	#endif // WIN32

	return oldestFileName;
}

std::string CFileIo::GetWorkingDirectory() {
	std::string directoryName = "";
	#ifdef _MSC_VER
	char result[MAX_PATH];
	directoryName = std::string(result, GetModuleFileName(NULL, result, MAX_PATH));
	return directoryName.substr(0, directoryName.rfind('\\'));
	#elif defined (__GNUC__)
	char result[PATH_MAX];
	size_t count = readlink("/proc/self/exe", result, PATH_MAX);
	directoryName = std::string(result, (count > 0) ? count : 0);
	return directoryName.substr(0, directoryName.rfind('/'));
	#endif
}

#ifdef _TESTING
#include "gtest\gtest.h"

TEST(ToString, ToStringFloat) {
	float test = 12345.6f;
	std::string temp = ToString(test);
	
	ASSERT_STREQ(temp.c_str(), "12345.6");
}

TEST(ToString, ToStringInt) {
	int test = -15;
	std::string temp = ToString(test);

	ASSERT_STREQ(temp.c_str(), "-15");
}

TEST(ToString, ToStringUInt) {
	unsigned int test = 798005;
	std::string temp = ToString(test);

	ASSERT_STREQ(temp.c_str(), "798005");
}

TEST(ToString, ToStringUShort) {
	unsigned short test = 64225;
	std::string temp = ToString(test);

	ASSERT_STREQ(temp.c_str(), "64225");
}

TEST(ToCamelCase, NoTokens) {
	std::string test = "helloworld";
	std::string result = CConvert::ToCamelCase(test);

	ASSERT_STREQ(test.c_str(), result.c_str());
}

TEST(ToCamelCase, SpaceTokens) {
	std::string test1 = "hello world";
	std::string result = CConvert::ToCamelCase(test1);

	ASSERT_STREQ(result.c_str(), "helloWorld");
	std::string test2 = "hello world goodbye earth";
	result = CConvert::ToCamelCase(test2);

	ASSERT_STREQ(result.c_str(), "helloWorldGoodbyeEarth");
}

TEST(ToCamelCase, UnderscoreTokens) {
	std::string test1 = "hello_world";
	std::string result = CConvert::ToCamelCase(test1);

	ASSERT_STREQ(result.c_str(), "helloWorld");
	std::string test2 = "hello_world_goodbye_earth";
	result = CConvert::ToCamelCase(test2);

	ASSERT_STREQ(result.c_str(), "helloWorldGoodbyeEarth");
}

TEST(UndoCamelCase, UnderscoreLowerCase) {
	std::string test1 = "helloWorld";
	std::string result = CConvert::UndoCamelCase(test1, '_', true);
	ASSERT_STREQ(result.c_str(), "hello_world");

	std::string test2 = "helloWorldGoodbyeEarth";
	result = CConvert::UndoCamelCase(test2, '_', true);
	ASSERT_STREQ(result.c_str(), "hello_world_goodbye_earth");
}



#endif // _TESTING

